/*
Copyright 2011. All rights reserved.
Institute of Measurement and Control Systems
Karlsruhe Institute of Technology, Germany

Authors: Andreas Geiger

libicp is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or any later version.

libicp is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
libicp; if not, write to the Free Software Foundation, Inc., 51 Franklin
Street, Fifth Floor, Boston, MA 02110-1301, USA 
*/

#include "ICP/icp.h"

using namespace std;

Icp::Icp (double *M,const int32_t M_num,const int32_t dim) :
  dim(dim), sub_step(10), max_iter(200), min_delta(1e-6) {
  
  // check for correct dimensionality
  if (dim!=2 && dim!=3) {
    cout << "ERROR: LIBICP works only for data of dimensionality 2 or 3" << endl;
    M_tree = 0;
    return;
  }
  
  // check for minimum number of points
  if (M_num<5) {
    cout << "ERROR: LIBICP works only with at least 5 model points" << endl;
    M_tree = 0;
    return;
  }

  // copy model points to M_data
  M_data.resize(boost::extents[M_num][dim]);
  for (int32_t m=0; m<M_num; m++)
    for (int32_t n=0; n<dim; n++)
      M_data[m][n] = (float)M[m*dim+n];

  // build a kd tree from the model point cloud
  M_tree = new kdtree::KDTree(M_data);
}

Icp::~Icp () {
  if (M_tree)
    delete M_tree;
}

void Icp::fit (double *T,const int32_t T_num,Matrix &R,Matrix &t,const double indist) {
  
  // make sure we have a model tree
  if (!M_tree) {
    cout << "ERROR: No model available." << endl;
    return;
  }
  
  // check for minimum number of points
  if (T_num<5) {
    cout << "ERROR: Icp works only with at least 5 template points" << endl;
    return;
  }
  
  // coarse matching
  vector<int32_t> active;
  active.clear();
  for (int32_t i=0; i<T_num; i+=sub_step)
    active.push_back(i);
  fitIterate(T,T_num,R,t,active);
  
  // fine matching
  if (indist<=0) {
    active.clear();
    for (int32_t i=0; i<T_num; i++)
      active.push_back(i);
  } else {
    active = getInliers(T,T_num,R,t,indist);
  }
  fitIterate(T,T_num,R,t,active);
}

void Icp::fitIterate(double *T,const int32_t T_num,Matrix &R,Matrix &t,const std::vector<int32_t> &active) {
  
  // check if we have at least 5 active points
  if (active.size()<5)
    return;
  
  // iterate until convergence
  for (int32_t iter=0; iter<max_iter; iter++)
  {
      if (fitStep(T,T_num,R,t,active)<min_delta)
          break;
//      cout << "iter " << iter << " translation " << t << "\n";
  }
}

void Icp::findKclosestPoints( double *T, int32_t T_num, Matrix &R, Matrix &t,
                             std::vector< std::pair<float, std::pair< Vec3Df, Vec3Df > > > & neighbors)
{
    vector<int32_t> active;
    for (int32_t i=0; i<T_num; i++)
      active.push_back(i);
    // kd tree query + result
    std::vector<float>         query(dim);
    kdtree::KDTreeResultVector result;

    // extract matrix and translation vector
    double r00 = R.val[0][0]; double r01 = R.val[0][1]; double r02 = R.val[0][2];
    double r10 = R.val[1][0]; double r11 = R.val[1][1]; double r12 = R.val[1][2];
    double r20 = R.val[2][0]; double r21 = R.val[2][1]; double r22 = R.val[2][2];
    double t0  = t.val[0][0]; double t1  = t.val[1][0]; double t2  = t.val[2][0];

    for (int32_t i=0; i<active.size(); i++)
    {
        // get index of active point
        int32_t idx = active[i];

        // transform point according to R|t
        query[0] = r00*T[idx*3+0] + r01*T[idx*3+1] + r02*T[idx*3+2] + t0;
        query[1] = r10*T[idx*3+0] + r11*T[idx*3+1] + r12*T[idx*3+2] + t1;
        query[2] = r20*T[idx*3+0] + r21*T[idx*3+1] + r22*T[idx*3+2] + t2;

        M_tree->n_nearest(query,1,result);
        // model point
        Vec3Df r = Vec3Df( M_tree->the_data[ result[0].idx][0],
                           M_tree->the_data[ result[0].idx][1],
                           M_tree->the_data[ result[0].idx][2]);
        neighbors.push_back(make_pair(result[0].dis , make_pair( Vec3Df( query[0], query[1], query[2]), r)));
    }
}
