/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef USERSTROKE_H
#define USERSTROKE_H
#include <vector>
#include "Vec3D.h"
#include <cfloat>
#include <QColor>

/**
 * @brief The UserStroke class
 */
class UserStroke
{
public:
    struct strokePoints{
        Vec3Df p;
        float r;
        bool newStroke;
    };

    UserStroke();
    virtual ~UserStroke(){}

    /**
     * @brief draw draw the stroke in the given color
     * @param color
     */
    void draw(std::vector<float> & color);

    /**
     * @brief addPoint add the point p of influence r in the stroke
     * @param p point position
     * @param r point radius
     */
    void addPoint(Vec3Df p, float r);

    /**
     * @brief deletePoint delete the point p
     * @param p point position
     * @param r point radius
     */
    void deletePoint(Vec3Df p, float r);

    /**
     * @brief deleteStrokeUnderPoint
     * @param point
     */
    void deleteStrokeUnderPoint(Vec3Df point);

    /**
     * @brief deleteAllPoints dekete all the points of the stroke
     */
    void deleteAllPoints();

    /**
     * @brief lastStrokeDrawn
     * function called when the mouse is released
     * set startNewStroke to true
     */
    void lastStrokeDrawn();

    /**
     * @brief isStarted
     * @return true if a stroke is being created
     */
    bool isStarted() { return !startNewStroke;}

    /**
     * @brief setFinished
     * ends a stroke
     */
    void setFinished() { startNewStroke = true; }

    /**
     * @brief getPoints
     * @return the stroke points
     */
    const std::vector<strokePoints> & getPoints(){return points;}

    /**
     * @brief getBGPoints
     * @return the background points
     */
    const std::vector<strokePoints> & getBGPoints(){return BGpoints;}

    /**
     * @brief getDeletedPoints
     * @return the deleted points
     */
    const std::vector<strokePoints> & getDeletedPoints(){return pointsBeingDeleted;}

    /**
     * @brief getNumberOfStrokes
     * @return  the number of points
     */
    int getNumberOfStrokes(){return numberOfStrokes;}

    /**
     * @brief getLastPoint
     * @return  the last point
     */
    Vec3Df getLastPoint(){
        return points[points.size()-1].p;
    }

    /**
     * @brief setColor
     * @param color set the stroke color
     */
    void setColor(QColor color)
    {
        strokeColor[0] =(float)(color.red()) / 255.f;
        strokeColor[1] =(float)(color.green()) / 255.f;
        strokeColor[2] =(float)(color.blue()) / 255.f;
    }

private:
    bool startNewStroke; /**< true if a new stroke is started */
    std::vector<Vec3Df> lastStroke;  /**< points of the last stroke drawn */
    std::vector < strokePoints > points; /**< vector of strokes */
    std::vector <strokePoints> BGpoints; /**< vector of background strokes */
    std::vector < bool > newStroke; /**< newStroke[i] == true if we are on a new stroke */
    std::vector<strokePoints> pointsBeingDeleted; /**< points being deleted */
    std::vector< float > strokeRadius; /**< stroke radius */
    int numberOfStrokes; /**< number of strokes */
    float strokeColor[3]; /**< stroke color */
};

#endif // USERSTROKE_H
