/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <vector>
#include "Vertex.h"
#include "Triangle.h"
#include "Edge.h"
#include "float.h"
#include "GLUtilityMethods.h"

/**
 * @brief The Mesh class
 */
class Mesh {

public:
    Mesh () {}
    /**
     * @brief Mesh Mesh constructor with a list of Vertice and Triangles
     * @param v vector of vertices
     * @param t vector of triangles
     */
    inline Mesh (const std::vector<Vertex> & v, 
                 const std::vector<Triangle> & t) : vertices (v), triangles (t)  {}
    /**
     * @brief Mesh Mesh constructor
     * @param mesh
     */
    inline Mesh (const Mesh & mesh) : vertices (mesh.vertices), triangles (mesh.triangles) {}
    virtual ~Mesh () {}

    /**
     * @brief clear call clearGeometry and clearTopology
     */
    void clear ();

    /**
     * @brief clearGeometry clear vertices
     */
    void clearGeometry ();

    /**
     * @brief clearTopology clear triangles
     */
    void clearTopology ();

    //**************************************************
    //                  ACCESSORS
    //**************************************************
    /**
     * @brief getVertices
     * @return return a vector of vertices
     */
    std::vector<Vertex> & getVertices () { return vertices; }

    /**
     * @brief getVertices
     * @return return a const vector of vertices
     */
    const std::vector<Vertex> & getVertices () const { return vertices; }

    /**
     * @brief getTriangles
     * @return return a vector of triangles
     */
    std::vector<Triangle> & getTriangles () { return triangles; }

    /**
     * @brief getTriangles
     * @return return a const vector of triangles
     */
    const std::vector<Triangle> & getTriangles () const { return triangles; }

    /**
     * @brief getOneRing returns the neighbors vertices of each vertex
     * @return  the neighbors vertices of each vertex
     */
    const std::vector< std::vector<unsigned int> >  & getOneRing() const {return oneRing;}

    /**
     * @brief getOneRingTr returns the neighbors triangles of each vertex
     * @return the neighbors triangles of each vertex
     */
    const std::vector< std::vector<unsigned int> >  & getOneRingTr() const {return oneRingTr;}

    /**
     * @brief getFaceNeighbors
     * @return returns the faces neighbors of each face
     */
    const std::vector< std::vector <unsigned int> > & getFaceNeighbors () const { return faceNeighbors; }
    /**
     * @brief getBBoxMax
     * @return return BBMax
     */
    const Vec3Df & getBBoxMax() const{return BBMax;}

    /**
     * @brief getBBoxMin
     * @return return BBmin
     */
    const Vec3Df & getBBoxMin() const{return BBMin;}

    /**
     * @brief getAverageEdgeHalfSize
     * @return return the average half size of all the edges
     */
    float getAverageEdgeHalfSize() const{return average_edge_halfsize;}

    /**
     * @brief getVerticeConnectedCompInd
     * @return returns the indice of the connected component of each vertex
     */
    const std::vector<unsigned int> & getVerticeConnectedCompInd() const {return vertexConnectedComp;}

    /**
     * @brief getConnectedCompRepresentatives
     * @return a representative vertex indice for each connected component
     */
    const std::vector<unsigned int> & getConnectedCompRepresentatives() const {
        return connectedCompRepresentatives;
    }

    /**
     * @brief getVertexIsOnShapeBoundary
     * @return returns a vector set to true if the vertex i is on the boundary of the shape
     */
    const std::vector<bool> & getVertexIsOnShapeBoundary() const {return vertexIsOnShapeBoundary;}

    Vec3Df getBarPosition( unsigned int t, double d0, double d1, double d2)
    {
        Vec3Df p0 = vertices[triangles[t].getVertex(0)].getPos();
        Vec3Df p1 = vertices[triangles[t].getVertex(1)].getPos();
        Vec3Df p2 = vertices[triangles[t].getVertex(2)].getPos();
        return p0*d0+d1*p1+d2*p2;
    }

    //**************************************************
    //           COMPUTE MESH PROPERTIES
    //**************************************************
    /**
     * @brief computeMeshProperties
     * general function to call at the beginning to compute several mesh properties
     * oneRing, oneRingTr, faceNeighbors, normals, and smoothed Normals, connected components
     */
    void computeMeshProperties();

    /**
     * @brief getBar
     * @param t indice of the triangle
     * @return the barycenter of the triangle t
     */
    Vec3Df getBar(int t);

    /**
     * @brief getNorm
     * @param t
     * @return return the normal of the triangle
     */
    Vec3Df getNorm(int t);

    /**
     * @brief computeVerticeSmoothNormals
     *  set the smooth normal of each vertex to :
     *  Sum_(v_nei)( exp(dist/0.245) * n_v_nei) || Sum_(v_nei)( exp(dist/0.245) * n_v_nei) ||
     */
    void computeVerticeSmoothNormals();

    /**
     * @brief recomputeSmoothVertexNormals set the normal of each vertex
     * @param normWeight
     *      case normWeight = 1 --> area weight
     *      case normWeight = 2 --> angle weight
     *      else                --> uniform weight
     * UniformWeight : the normal of each vertex is equal to the uniform sum of the normals of each neighbors faces
     * AreaWeight    : the sum is weighted by the area of each triangle
     * AngleWeight   : the sum is weighted by the angle formed by the triangle at vertex v
     */
    void recomputeSmoothVertexNormals (unsigned int normWeight);

    /**
     * @brief computeTriangleNormals
     */
    void computeTriangleNormals ();

    /**
     * @brief smoothTriangleNormals compute the smoothNormal vector of each triangle to:
     *  Sum_(tr_nei)( exp(dist/0.245) * n_tr_nei) || Sum_(tr_nei)( exp(dist/0.245) * n_tr_nei) ||
     */
    void smoothTriangleNormals();

    /**
     * @brief computeBoundingBox
     */
    void computeBoundingBox();

    /**
     * @brief computeAveragePosAndRadius
     * @param center center of the mesh
     * @param radius radius of a bounding cercle
     */
    void computeAveragePosAndRadius ( Vec3Df & center, float & radius);

    /**
     * @brief computeAverageEdgeHalfSize
     */
    void computeAverageEdgeHalfSize();

    /**
     * @brief findShapeBoundaries find the boundaries of the mesh (if any)
     * @param vertexIsOnShapeBoundary vector of boolean. Set to true if a vertex is on a boundary
     */
    void findShapeBoundaries(std::vector<bool> &vertexIsOnShapeBoundary );

    /**
     * @brief computeGaussianCurvature
     * @param gaussianCurvature
     */
    void computeGaussianCurvature( std::vector<float> & gaussianCurvature);

    /**
     * @brief findConnectedComponents generic function to find connected components
     * @param parts parts[i]=connected_component_indice of elements[i]
     * @param representatives vectors containing  1 representative element for each connected comp
     * @param elements can be either faces or vertice
     * @param neighbors neighbors is either faceNeighbors of verticesNeighbors
     */
    void findConnectedComponents( std::vector<unsigned int> & parts,
                                  std::vector< unsigned int > & representatives,
                                  std::vector<unsigned int> &  elements,
                                  const std::vector<std::vector<unsigned int > > & neighbors);

    /**
     * @brief vertexCloseToPoint
     * @param point
     * @param radius the distance between p and the vertex found has to be < radius.
     * @return the closest vertex to the point p
     */
    int vertexCloseToPoint(Vec3Df point, float radius);

    /**
     * @brief triangleCloseToPoint
     * @param point
     * @param radius the distance between p and the triangle barycenter has to be < radius.
     * @return the triangle that has the closest barycenter to the point p
     */
    int triangleCloseToPoint(Vec3Df point, float radius);

    /**
     * @brief barycentricCoordinates
     * @param point 3D point we look at (in)
     * @param t indice of the triangle on which is the point
     * @param d0 barycentric coordinate of p
     * @param d1
     * @param d2
     * @return true if the point is projected on the mesh
     */
    bool barycentricCoordinates(Vec3Df point, unsigned int & t,
                                double & d0, double & d1, double &d2);

    void update();
    void draw();


protected:

    /**
     * @brief triangleArea
     * @param i
     * @return the area of triangle i
     */
    float triangleArea(unsigned int i);

    /**
     * @brief computeVoronoiArea
     * @param v
     * @return voronoi area of vertex v
     */
    float computeVoronoiArea(unsigned int v);

    /**
     * @brief computeOneRing
     * fill the oneRing vector with all the neighbors vertices of each vertex
     */
    void computeOneRing();

    /**
     * @brief computeOneRingTr
     * fill the oneRingTr vector with all the neighbor faces of each vertex
     */
    void computeOneRingTr();

    /**
     * @brief findFacesNeighbors
     * fill the faceNeighbors vector  with all the neighbor faces of each face
     */
    void findFacesNeighbors();

    std::vector<Vertex> vertices; /**< vector of vertex */
    std::vector<Triangle> triangles; /**< vector of faces */

    std::vector< std::vector<unsigned int> > oneRing; /**< neighbor vertices for each vertex */
    std::vector< std::vector<unsigned int> > oneRingTr; /**< neighbor faces for each vertex */
    std::vector< std::vector <unsigned int> > faceNeighbors; /**< neighbor faces for each face */

    double average_edge_halfsize; /**< average half size of the edges */

    Vec3Df BBMax, BBMin; /**< bounding box min and max */

    std::vector<unsigned int> connectedCompRepresentatives;
    /**< contains 1 representative vertex for each component */

    std::vector<unsigned int> vertexConnectedComp;
    /**< connected component indice for each vertex */

    std::vector<bool> vertexIsOnShapeBoundary;
    /**< true if the vertex i is on the shape boundary */
};




// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
