/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef COMPSELECTION_H
#define COMPSELECTION_H
#include "Selection.h"
#include "Histogram2D.h"

/**
 * @brief The componentShapeContext struct
 * contains the indice of the component,
 * it center and maximum radius (find by a miniball algorithm)
 * it descriptor (here a ASC, 2D histogram)
 * the distance and rank of the comp from the ref
 */
struct componentShapeContext{
    unsigned int _indice;
    Vec3Df _center;
    float _radiusMax;
    Hist2Df _shapeContext;
    float _nDist;
    unsigned int _rankToRef;
};

/**
 * @class connected components selection
 */
class CompSelection : public Selection
{
public:
    CompSelection():Selection(){}

    ~CompSelection(){}

    /**
     * @brief init initialize the connected components selection
     *  compute the ASC for each component
     * @param m mesh
     */
    void init(const Mesh &m);

    /**
     * @brief clear
     */
    void clear();


    //***************************************************
    //                  SELECTION
    //***************************************************

    /**
     * @brief startSelection
     * @param stroke user stroke (here just a click is used)
     */
    void startSelection(UserStroke & stroke);

    /**
     * @brief updateSelection
     * @param stroke last user click is used to define the selected component
     */
    void updateSelection(UserStroke & stroke);

    /**
     * @brief endSelection
     * @param stroke not used
     */
    void endSelection(UserStroke &stroke);

    /**
     * @brief eraseSelection
     * @param stroke
     */
    void eraseSelection(UserStroke & stroke);

    /**
     * @brief clearSelections
     */
    void clearSelections();

    /**
     * @brief refineSelection (do nothing for components)
     * @param stroke
     */
    void refineSelection(UserStroke & stroke){
        Q_UNUSED(stroke);
    }
    //***************************************************
    //                  EXPANSION
    //***************************************************
    /**
     * @brief findSimilarSelections
     * set all the components with a smaller rank than the nb_of_similar_sel as selected
     * @param stroke not used
     */
    void findSimilarSelections(UserStroke& stroke);


    /**
     * @brief changeSimiliarityThreshold
     * set nb_of_similar_sel = val and recompute the similar selections selected
     * @param val
     */
    void changeSimiliarityThreshold(int val);

    /**
     * @brief getRefHistogram
     * @return the descriptor of the reference component
     */
    QPixmap getRefHistogram() {
        return _shapeContexts[componentSelected]._shapeContext.drawHistogram();
    }

    /**
     * @brief getUnderMouseHistogram
     * @param v vertex under the mouse
     * @param dist distance between the reference descriptor and the one under the mouse
     * @return the descriptor of the component under the mouse
     */
    QPixmap getUnderMouseHistogram(unsigned int v, float & dist);

    /**
     * @brief getSimilarityDist
     * @return
     */
    const std::vector<float> & getSimilarityDist(){
        normalized_similarity_field.clear();
        for (unsigned int i = 0; i < vertices.size(); i++)
            normalized_similarity_field.push_back(_shapeContexts[vertexConnectedComp[i]]._nDist);
        return normalized_similarity_field;
    }

    /**
     * @brief drawShapeContext
     * @param i
     */
    void drawShapeContext(unsigned int i);

    /**
     * @brief saveTimings
     * @param filename
     */
    void saveTimings(const QString &filename);

protected:
    /**
     * @brief computeASCs compute the descriptors (approximate shape context)
     * of all connected components
     */
    void computeASCs();

    /**
     * @brief computeASC compute the descriptor of current component
     * @param curComp
     */
    void computeASC(componentShapeContext & curComp );

    /**
     * @brief recomputeSimilarityDist compute distances between reference and others components
     */
    void recomputeSimilarityDist();

    /**
     * @brief set all the components with a smaller rank than the nb_of_similar_sel as selected
     */
    void findSimilarSelections();

private:
    std::vector<float> normalized_similarity_field;
    /**< to display the distance between components (values between 0 and 1)*/
    std::vector<componentShapeContext> _shapeContexts;
    /**< all the infos for each connect component are here */
    unsigned int componentSelected; /**< selected component indice */

};

#endif // COMPSELECTION_H
