/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PARTSELECTION_H
#define PARTSELECTION_H
#include "Histogram1D.h"

#include "Selection.h"
#include "CholmodStruct.h"

/**
 * @brief The PartSelection class
 */
class PartSelection: public Selection
{
    /**
     * @brief The potentialSelection struct
     * contains the descriptor hist
     * the distance between the potential selection and the reference
     * the range of the potential selection PartSelection
     * all the faces in this potential selections
     * the sets of c0 and c1 constraints (vertices indices)
     */
    struct potentialSelection{
        Hist1Df hist;
        float dist_to_ref;
        unsigned int range;
        std::vector<unsigned int> faces;
        std::vector<unsigned int> c0;
        std::vector<unsigned int> c1;
    };

public:
    PartSelection():Selection(){}
    ~PartSelection(){}

    /**
     * @brief init initialize the selection from the mesh m
     * compute the laplacian, the conformal factor, initialize cholmod
     * and set severals members to initialized values
     * @param m
     */
    void init(const Mesh &m);

    /**
     * @brief clear
     */
    void clear();

    /**
     * @brief saveTimings
     * @param filename
     */
    void saveTimings(const QString &filename);
    //***************************************************
    //                  SELECTION
    //***************************************************

    /**
     * @brief startSelection the user stroke is used to define several constraints
     * these constraints allows to define a userHarmonicField
     * @param stroke
     */
    void startSelection(UserStroke & stroke);

    /**
     * @brief updateSelection constraints are added and the harmonic user field is updated
     * each time the stroke is growing
     * @param stroke
     */
    void updateSelection(UserStroke & stroke);

    /**
     * @brief endSelection constraints are added and the harmonic user field is updated
     * @param stroke
     */
    void endSelection(UserStroke &stroke);

    /**
     * @brief eraseSelection remove the triangles under the user stroke from the selection
     * @param stroke
     */
    void eraseSelection(UserStroke & stroke);

    /**
     * @brief refineSelection add the triangles under the user stroke to the selection
     * @param stroke
     */
    void refineSelection(UserStroke & stroke);
    /**
     * @brief clearSelections
     */
    void clearSelections();

    //***************************************************
    //                  EXPANSION
    //***************************************************

    /**
     * @brief findSimilarSelections
     * find the k most similar selections from the reference one
     * use the conformal factor as a similarity map to retrieve potential part boundaries
     * specify constraints along conformal factor isolines and apply the part selection process
     * for each potential part, compute it similarity to the reference and rank it
     * @param stroke
     */
    void findSimilarSelections(UserStroke& stroke);

    /**
  * @brief changeSimiliarityThreshold
  * @param val
  * set the similarity threshold to val and recompute the similar selection
  */
    void changeSimiliarityThreshold(int val);

    /**
     * @brief getConformalFactor
     * @return the conformal factor map
     */
    const std::vector<float> & getConformalFactor()const {return conformalFactor;}

    /**
     * @brief getUserField
     * @return the harmonic field defined using the user strokes
     */
    const std::vector<float> & getUserField();

    /**
     * @brief drawUserConstraints
     */
    void drawUserConstraints();

    /**
     * @brief drawSimilarConstraints
     */
    void drawSimilarConstraints();

    /**
     * @brief drawIsolines
     */
    void drawIsolines();

    /**
     * @brief getRefHistogram
     * @return
     */
    QPixmap getRefHistogram() { QPixmap ret;
                                if (potentialSelections.empty())
                                    return ret;
                                return refHist.drawHistogram();}

    /**
     * @brief getUnderMouseHistogram
     * @param v
     * @param dist
     * @return
     */
    QPixmap getUnderMouseHistogram(unsigned int v, float &dist);

    /**
     * @brief getCotanLaplacian
     * @return
     */
    CotanLaplacian* getCotanLaplacian(){
        return _cotanLaplacian;
    }
protected:

    void findSelectedFaces();

    /**
     * @brief findTrianglesInsideRadius
     * find all the triangles under the sphere (center, radius) which doesn't have the inside_value
     * and set their inside value to new_inside_value
     * @param point sphere center
     * @param radius sphere radius
     * @param inside
     * @param inside_value previous inside value
     * @param new_inside_value new inside value
     * @return true if any triangle is found and changed
     */
    bool findTrianglesInsideRadius( Vec3Df point, float radius,
                                    std::vector<int> &  inside,
                                    int inside_value,
                                    int new_inside_value);

    //***************************************************
    //                  ISOLINE STUFFS
    //***************************************************
    /**
     * @brief findIsolines find the isolines k closest isolines
     * that go through to the user strokes
     * @param curStroke
     */
    void findIsolines(UserStroke & curStroke);

    /**
     * @brief findIsoline generic function to retrieve isolines
     * @param refIsoVal (in) value of the wanted isolines
     * @param isoValues (in) scalar field on the vertices
     * @param isoLines_i (out) return all the isolines that have the refIsoVal value
     */
    void findIsoline(float refIsoVal, const std::vector<float> & isoValues,
                     std::vector< std::vector<Vec3Df> > & isoLines_i );

    /**
     * @brief findIsolineToKeep if we have several isolines with the same isovalue
     * we just want to keep the one that is the closest to the user stroker
     * @param currStroke user stroke
     * @param isoLines_i vector of isolines
     * @return indice of the isoline to keep
     */
    unsigned int findIsolineToKeep(UserStroke & currStroke,
                                   std::vector< std::vector<Vec3Df> > & isoLines_i);

    /**
     * @brief findBestIsoline
     * choose the best isoline, such as the algorithm described in
     * "Mesh Decomposition using cross boundary brushes" (Zheng, 2010)
     */
    void findBestIsoline();

    /**
     * @brief findNextIsolineElement follow the path of one isoline
     * @param isoline points of the current isoline we are following
     * @param triangleViewed set to true if the triangle is already treated
     * @param triangleInd values of the triangle depending on the way the isoline traverse it (from 0 to 7)
     * see PartSelection.cpp -->findIsoline for more details
     * @param isoValues
     * @param refIsoVal
     * @param v0
     * @param v1
     * @param t
     */
    void findNextIsolineElement( std::vector<Vec3Df> & isoline,
                                 std::vector<bool> & triangleViewed,
                                 const std::vector<unsigned int> & triangleInd,
                                 const std::vector<float> & isoValues,
                                 const float refIsoVal,
                                 unsigned int v0,
                                 unsigned int v1,
                                 unsigned int t);

    //***************************************************
    //                  EXPANSION
    //***************************************************

    /**
     * @brief computeConformalFactor
     * such as "the paper "Characterizing shapes using conformal factors" (Ben-Chen, 2008)
     */
    void computeConformalFactor();

    /**
     * @brief findConformalIsolines
     * find all the isolines that have the same isovalue than
     * the average of the isovalues of the user stroke
     * @param strokeInit (in) user stroke
     * @param conf_iso_lines (out) vector of isolines
     */
    void findConformalIsolines( const std::vector<UserStroke::strokePoints> & strokeInit,
                                std::vector< std::vector<Vec3Df> >  & conf_iso_lines );


    /**
     * @brief findMultipleSelectedFaces find the faces that are in the k° potential selection
     * change the faceInSelectionK vector accordingly
     * @param curSelection k° potential selection
     * @return false if the k° potential selection is not selected or is the reference selection
     * true otherwise
     */
    bool findMultipleSelectedFaces(const potentialSelection & curSelection);

    /**
     * @brief setPotentialSelection find the faces that are part of the potential selection
     * compute the potential selection descriptor, and distance to the reference
     * all these data are stored in the curSel structure
     * @param curSel potential selection
     */
    void setPotentialSelection( potentialSelection & curSel );

    /**
     * @brief recomputeSimilarityDist
     */
    void recomputeSimilarityDist();
    //***************************************************
    //      FNS FOR THE COMPUTATION OF HARMONIC FIELDS
    //***************************************************
    /**
     * @brief setHarmonicConstraints
     * set the P_0 and P_1 constraints to build the user harmonic field
     * @param nb_of_ref_points if you have to many points you can add a constraint only for some of them
     * @param strokePoints points of the stroke (virtual or from the user)
     * @param constraints0 (out) set of 0 constraints found
     * @param constraints1 (out) set of 1 constraints found
     */
    void setHarmonicConstraints( unsigned int nb_of_ref_points,
                                 const std::vector<Vec3Df> & strokePoints,
                                std::vector<unsigned int> & constraints0,
                                std::vector<unsigned int> & constraints1 );

    /**
     * @brief setHarmonicConstraints
     * create a vector with the 3D positions of the stroke points and call setHarmonicConstraints
     * @param nb_of_ref_points
     * @param strokePoints
     * @param constraints0
     * @param constraints1
     */
    void setHarmonicConstraints( unsigned int nb_of_ref_points,
                                 const std::vector<UserStroke::strokePoints> & strokePoints,
                                std::vector<unsigned int> & constraints0,
                                std::vector<unsigned int> & constraints1 );

    // CHOLMOD stuff
    /**
     * @brief initCholmod initialize the cholmod structure to solve the linear system
     * @param constraints true if there is any constraint
     * @param constraints0 0-constraints found from the user interaction or the virtual strokes
     * @param constraints1 1-constraints found from the user interaction or the virtual strokes
     */
    void initCholmod(bool constraints,
                     std::vector<unsigned int> & constraints0,
                     std::vector<unsigned int> & constraints1 );

    /**
     * @brief setCholmodUpdate
     * @param update if true : we need to update the factorization,
     *                          and solve the system with the new constraints
     *               if false : we only need to downdate the factorization
     * @param _constraints0 new 0-constraints if update
     * @param _constraints1 new 1-constraints if update
     */
    void setCholmodUpdate( bool update,
                           std::vector<unsigned int> & _constraints0,
                           std::vector<unsigned int> & _constraints1  );


private:
    CholmodLSStruct cholmod_struct; /**< cholmod structure used to compute harmonic fields*/
    MeshLaplacian* _laplacian; /**< mesh laplacian */
    CotanLaplacian* _cotanLaplacian; /**< mesh laplacian with cotan weightd*/
    bool cholmodInitialized; /**< true if cholmod is already iniatialized*/

    // To find Similar Parts
    std::vector<float> conformalFactor; /**< conformal factor field*/
    std::vector< std::vector<Vec3Df> > potentialStrokes; /**< vector of potential strokes*/
    std::vector <potentialSelection> potentialSelections; /**< vector of potential selections
(see the potential selection structure for details)*/
    Hist1Df refHist; /**< conformal factor distribution of the reference selection */

    // user constraints
    std::vector<unsigned int> constraints0; /**< 0-constraints defined by the user stroke */
    std::vector<unsigned int> constraints1; /**< 1-constraints defined by the user stroke */

    std::vector<std::vector <Vec3Df> > isolines; /**< isolines found from the user harmonic field */
    unsigned int refIsoline; /**< 0-constraints defined by the user stroke */
    std::vector<float> userDefinedHarmonicField;
    std::vector<float> normalizedHarmonicField;
    std::vector< std::vector< std::pair<unsigned int, float> > > timing_by_part;
};

#endif // PARTSELECTION_H
