/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef HISTOGRAM2D_H
#define HISTOGRAM2D_H

#include "Debug.h"
#include <iostream>
#include <climits>
#include <vector>
#include <cfloat>
#include <QColor>
#include <QPixmap>
#include <QPainter>
#include <assert.h>
#include <fstream>

template<typename T, typename histT>

class Histogram2D
{
public:
    Histogram2D()
    {
        xbin_nb = 0;
        ybin_nb = 0;
        bounds[0] = bounds[2] = (T) FLT_MAX;
        bounds[1] = bounds[3] = (T) FLT_MIN;
    }

    Histogram2D( unsigned int x_bin_nb, unsigned int y_bin_nb,
                 T  x_bin_min, T x_bin_max,  T y_bin_min, T y_bin_max )
    {
        xbin_nb = x_bin_nb;
        ybin_nb = y_bin_nb;

        std::vector< histT> zeroH;
        for (unsigned int j = 0; j < ybin_nb; j++)
            zeroH.push_back((histT) 0.f);
        hist.clear();
        for (unsigned int i = 0; i < xbin_nb; i++)
            hist.push_back(zeroH);

        bounds[0] = x_bin_min;
        bounds[1] = x_bin_max;
        bounds[2] = y_bin_min;
        bounds[3] = y_bin_max;
    }

    void setBinsNumber( unsigned int x_nb, unsigned int y_nb)
    {
        xbin_nb = x_nb;
        ybin_nb = y_nb;
        hist.resize(xbin_nb);
        for (unsigned int i = 0; i < xbin_nb; i++)
            hist[i].resize(ybin_nb);
    }

    unsigned int getBinsNumber()
    {
         return xbin_nb* ybin_nb;
    }

    void setBounds(T  x_bin_min, T x_bin_max,  T y_bin_min, T y_bin_max )
    {
        bounds[0] = x_bin_min;
        bounds[1] = x_bin_max;
        bounds[2] = y_bin_min;
        bounds[3] = y_bin_max;
    }

    void addValue ( T x_val, T y_val, histT value)
    {
        unsigned int x = (x_val-bounds[0]) / (bounds[1]-bounds[0]) * (T) (xbin_nb-1);
        unsigned int y = (y_val-bounds[2]) / (bounds[3]-bounds[2]) * (T) (ybin_nb-1);
        assert( (x < xbin_nb) && (y < ybin_nb));
//        if (y_val == 0)
//            std::cout << " y nul " << y<< "\n";
        hist[x][y] += value;
    }

    void normalize()
    {
        histT sum = 0.f;
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for (unsigned int j =0; j < ybin_nb; j++)
            {
                sum+= hist[i][j];
            }
        }

        if (sum!=0.f)
        {
            for (unsigned int i = 0; i < xbin_nb; i++)
            {
                for (unsigned int j = 0; j < ybin_nb; j++)
                {
                    hist[i][j]/=sum;
                }
            }
        }
    }

    void clearValues(){
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for (unsigned int j = 0; j < ybin_nb; j++)
            {
                hist[i][j]= 0;
            }
        }
    }

    histT distanceL2( const Histogram2D & h2 ){
        histT res = 0.f;
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for (unsigned int j = 0; j < ybin_nb; j++)
            {
                res += (hist[i][j]-h2.hist[i][j])* (hist[i][j]-h2.hist[i][j]);
            }
        }
        return sqrt(res);
    }

    histT distanceL1( const Histogram2D & h2 ){
        histT res = 0.f;
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for (unsigned int j = 0; j < ybin_nb; j++)
            {
                res += abs(hist[i][j]-h2.hist[i][j]);
            }
        }
        return sqrt(res);
    }

    histT distanceChi2(const Histogram2D & h2)
    {
        histT res = 0.f;
        assert( (xbin_nb == h2.xbin_nb) && (ybin_nb == h2.ybin_nb));
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for (unsigned int j = 0; j < ybin_nb; j++)
            {
                if ( hist[i][j]+ h2.hist[i][j] != 0 )
                {
                    res += (hist[i][j]-h2.hist[i][j])* (hist[i][j]-h2.hist[i][j])
                            /  (abs(hist[i][j]+h2.hist[i][j]));
                }
            }
        }
//        Trace(1, " res chi2 " << res/2.f);
        return res/2.f;
    }

    histT distanceChi2Normalized(const Histogram2D & h2)
    {
        histT sum1 = 0.f;
        histT sum2 = 0.f;
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for (unsigned int j = 0; j < ybin_nb; j++)
            {
                sum1+=hist[i][j];
                sum2+=h2.hist[i][j];
            }
        }

        histT res = 0.f;
        assert( (xbin_nb == h2.xbin_nb) && (ybin_nb == h2.ybin_nb));
        histT vh1, vh2;
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for (unsigned int j = 0; j < ybin_nb; j++)
            {
                vh1 = hist[i][j]/sum1;
                vh2 = h2.hist[i][j]/sum2;
                if ( vh1 + vh2 != 0.f)
                    res += (vh1-vh2)* (vh1-vh2) / (vh1+vh2);

            }
        }
//        Trace(1, " res chi2 " << res/2.f);
        return res/2.f;
    }

    void writeValues(std::ofstream & out )
    {
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for (unsigned int j = 0; j < ybin_nb; j++)
                out << " " << i*ybin_nb+j+1 <<":"<<hist[i][j];
        }
    }
    void getValues(std::vector<double> & out )
    {
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for (unsigned int j = 0; j < ybin_nb; j++)
                out.push_back((double) hist[i][j]);
        }
    }


    QPixmap drawHistogram(){
        // Create an histogram_ and display it
        const int height = 150;
        const int top_margin = 15;
        const int left_margin = 15;
        const int drawing_height = 120;
        const int cell_height = drawing_height/ybin_nb;
        const int width = 150;
        const int cell_width = (width - 2*left_margin)/xbin_nb;
//        const int text_margin = 10;
        const int text_height = 10;

        drawHist = QPixmap(width,height+text_height);
        drawHist.fill(Qt::white);

        QPainter painter(&drawHist);
        painter.setPen(Qt::black);
        painter.setBrush(Qt::white);
        //painter.setFont(QFont("Arial", 30));

        histT max_value = 0.f;
        histT min_value = FLT_MAX;
        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for (unsigned int j =0; j < ybin_nb; j++)
            {
                if (max_value < hist[i][j])
                    max_value = hist[i][j];
                if (min_value > hist[i][j])
                    min_value = hist[i][j];
            }
        }

        // colored histogram

        // draw
        int i_pos=left_margin;
        int j_pos=top_margin;
        histT max_min = max_value-min_value;
        histT val;

        for (unsigned int i = 0; i < xbin_nb; i++)
        {
            for ( int j = ybin_nb-1; j >= 0; j--)
            {
                if (max_min != 0)
                    val = (hist[i][j]-min_value)/max_min;
                else
                    val = hist[i][j];

                val*=255;
//                if (j == 0)
//                    Trace(1, " i" << i << "- val =" << val << " -max : "<< max_value << " min value " << min_value);
                QColor color(val*0.8, 255-val*0.8, 255-0.8*val);
                color.toHsv();
                color.setHsv(color.hue(), color.saturation()*3/4, color.value());
                painter.fillRect(i_pos, j_pos, cell_width, cell_height, color);
                j_pos += cell_height;
            }
            i_pos += cell_width;
            j_pos=top_margin;
        }

        painter.setPen(Qt::white);
        j_pos= top_margin;
        for (unsigned int j = 0; j < ybin_nb+1; j++)
        {
            painter.drawLine(QPoint(left_margin,j_pos),
                                    QPoint(left_margin+ static_cast<int>(cell_width*xbin_nb) ,
                                          j_pos));
            j_pos += cell_height;
        }
        painter.setPen(Qt::white);
        i_pos=left_margin;
        for (unsigned int i = 0; i < xbin_nb+1; i++)
        {
            painter.drawLine(QPoint(i_pos, top_margin),
                                    QPoint( i_pos ,
                                           top_margin + static_cast<int> (cell_height*ybin_nb) ));
            i_pos += cell_width;
        }


        painter.setPen(Qt::black);
        QRect min_text_rect (0, top_margin, left_margin, top_margin+cell_height*ybin_nb);
        painter.drawText(min_text_rect, Qt::AlignTop, "pi" );

        QRect min_text_rect1 (0, top_margin, left_margin, cell_height*ybin_nb);
        painter.drawText(min_text_rect1, Qt::AlignBottom, "0" );


        QRect min_text_rect2 (left_margin, cell_height*ybin_nb+top_margin, width, 15);
        painter.drawText(min_text_rect2, Qt::AlignLeft, "0" );

        QRect min_text_rect3 (left_margin, cell_height*ybin_nb+top_margin, width-15, 15);
        painter.drawText(min_text_rect3, Qt::AlignRight, "1" );

//        QRect max_text_rect (max_tr_pos - tr_width/2., tr_y, tr_width, text_height);
//        text.setNum(max,'f',2);
//        painter.drawText(max_text_rect, Qt::AlignCenter, text);

//        QRect max_nb_rect (max_nb_pos- tr_width/2., tr_y, tr_width, text_height);
//        text.setNum(max_nb_value,'f',2);
//        painter.drawText(max_nb_rect, Qt::AlignCenter, text);

        // draw bottom horizontal line
    //        painter.setPen(Qt::blue);

    //        painter.drawLine(QPoint(left_margin, drawing_height + top_margin),
    //                         QPoint(left_margin + static_cast<int>(size)*cell_width,
    //                                drawing_height + top_margin));


    //        // draw min value and max value

    //        const double max_nb_value = min + m_id*step;

        return drawHist;

    }


protected:
    std::vector< std::vector<histT> > hist;
    unsigned int xbin_nb;
    unsigned int ybin_nb;
    T bounds[4];
    QPixmap drawHist;


};
typedef Histogram2D<float, float> Hist2Df;

#endif // HISTOGRAM2D_H
