/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SVMCLASSIFICATION_H
#define SVMCLASSIFICATION_H
#include "svm.h"
#include <vector>
#include <Histogram2D.h>

/**
 * @brief The SVMClassification class
 * to handle the use of libSVM
 * Support Vector Machine algorithm
 */
class SVMClassification
{

public:
    SVMClassification();

    /**
     * @brief setSVMproblem
     * @param _labels labels of the datas (0 or 1)
     * @param _vectors datas expressed in the feature space
     */
    void setSVMproblem(std::vector<int> & _labels,
                       std::vector<std::vector<double> > & _vectors);


    // --------------------------------------------//
    // ------------- Set parameters ---------------//
    // --------------------------------------------//

    /**
     * @brief setSVMparameters
     * @param _svm_type type of the svm (default C-SVM)
     * @param _kernel_type type of the kernel function (default  RBF)
     * @param _degree degree in the kernel function (default 3)
     * @param _gamma gamma in the kernel function (defautl 1/num_features)
     * @param _coef0 coef0 in the kernel function (default 0)
     */
    void setSVMparameters(int _svm_type,
                          int _kernel_type,
                          int _degree,
                          double _gamma, double _coef0 );

    /**
     * @brief setSVMtype
     * @param _type can be C_SVC, NU_SVC, ONE_CLASS, EPSILON_SVR, NU_SVR
     */
    void setSVMtype(int _type) {param.svm_type = _type;}

    /**
     * @brief setKernelType
     * @param _kernel_type can be LINEAR, POLY, RBF, SIGMOID, PRECOMPUTED
     *          LINEAR:	u'*v
     *          POLY:	(gamma*u'*v + coef0)^degree
     *          RBF:	exp(-gamma*|u-v|^2)
     *          SIGMOID:	tanh(gamma*u'*v + coef0)
     *          PRECOMPUTED: kernel values in training_set_file
     */
    void setKernelType(int _kernel_type) {param.kernel_type = _kernel_type;}

    /**
     * @brief setPolynomialKernelDegree
     * @param _degree
     */
    void setPolynomialKernelDegree(int _degree) {param.degree = _degree;}

    /**
     * @brief setKernelGamma
     * @param _gamma
     */
    void setKernelGamma(double _gamma){param.gamma = _gamma;}

    /**
     * @brief setCoef0 for poly/sigmoid kernels only
     * @param _coef0
     */
    void setCoef0(double _coef0){param.coef0 = _coef0;}

    //           parameters for training only
    /**
     * @brief setCacheSize
     * @param _cacheSize the size of the kernel cache, specified in megabytes.
     */
    void setCacheSize(double _cacheSize){param.cache_size = _cacheSize;}

    /**
     * @brief setStoppingCriteria
     * @param _eps stopping criterion. (we usually use 0.00001 in nu-SVC, 0.001 in others)
     */
    void setStoppingCriteria(double _eps){param.eps = _eps;}

    /**
     * @brief setCparameter
     * @param _C cost of constraints violation. (for C_SVC, EPSILON_SVR, and NU_SVR)
     */
    void setCparameter(double _C){param.C = _C;}

    /**
     * @brief setWeightLabel
     * @param _weight_label
     * @param _weight
     */
    void setWeightLabel(std::vector<int> _weight_label, std::vector<double> _weight)
    {
        param.nr_weight = _weight_label.size();
        param.weight_label = (int *)realloc(param.weight_label,sizeof(int)*_weight_label.size());
        param.weight = (double *)realloc(param.weight,sizeof(double)*_weight_label.size());
        for (unsigned int i = 0; i < _weight_label.size(); i++)
        {
            param.weight_label[i] = _weight_label[i];
            param.weight[i]= _weight[i];
        }
        param_allocated = true;
    }

    /**
     * @brief setNu
     * @param _nu
     */
    void setNu(double _nu){param.nu = _nu;}

    /**
     * @brief setP
     * @param _p
     */
    void setP(double _p){param.p = _p;}

    /**
     * @brief setShrinking whether to use the shrinking heuristics,
     * @param _shrinking  0 or 1 (default 1)
     */
    void setShrinking(int _shrinking){param.shrinking = _shrinking;}

    /**
     * @brief setProbability whether to train a SVC or SVR model for probality estimates
     * @param _probability 0 or 1 (default 0)
     */
    void setProbability(int _probability){param.probability = _probability;}

    /**
     * @brief do_cross_validation
     */
    void do_cross_validation();
    // todo svm scale ?

    /**
     * @brief train train the svm on the labeled datas
     */
    void train();

    /**
     * @brief saveModelInFile
     * @param model_file_name
     */
    void saveModelInFile(std::string model_file_name);

    /**
     * @brief predicts the labels of vectors
     * @param _labels
     * @param _vectors
     */
    void predict(std::vector<int> & _labels,
                 std::vector<std::vector<double> > & _vectors);

    /**
     * @brief predictWithProb predicts the labels of vectors and
     * also return the probability associated to each predicted label
     * @param _labels
     * @param _vectors
     * @param probabilities
     */
    void predictWithProb (std::vector<int> & _labels,
                          std::vector<std::vector<double> > & _vectors,
                          std::vector< std::vector<double> > & probabilities);

    /**
     * @brief clear
     */
    void clear();
private:

    struct svm_parameter param; /**< parameters necessary for libsvm */
    struct svm_problem prob; /**< describes the problem */
    struct svm_model *model; /**< produced by svm_train */
    struct svm_node *x_space;
    int cross_validation;
    int nr_fold;
    bool model_allocated;
    bool param_allocated;
    bool x_space_allocated;
    bool prob_allocated;
};

#endif // SVMCLASSIFICATION_H
