/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "MainWindow.h"

#include "qdatetime.h"

MainWindow::MainWindow(int argc, char **argv)
{
    _debugMode = false;
    QWidget * selectWidget = new QWidget(this);
    QGridLayout * gridLayout = new QGridLayout(selectWidget);

    this->resize(800, 600);
    _viewer = new GLViewer(this);

    QSize s (800, 600);
    this->setMinimumSize (s);

    if (argc > 1) {
        string fullName;
        if (argv[1][0] == '/')
            fullName = string (argv [1]);
        else
            fullName = string (qPrintable (QDir::currentPath ())) + "/" + string (argv [1]);
        _viewer->open (QString (fullName.c_str ()));

    }
    gridLayout->addWidget(_viewer, 0, 1, 1, 1);

    connect(_viewer, SIGNAL(debugMode(bool)), this, SLOT(setDebug(bool)) );

    _selectionDock = new SelectionDock( _viewer, this);
    _selectionDock->setFeatures(QDockWidget::NoDockWidgetFeatures);
    QSize s2 (800, 105);
    _selectionDock->setMaximumSize(s2);
    _selectionDock->setMinimumSize(s2);

    this->addDockWidget(Qt::TopDockWidgetArea, _selectionDock);

    _visuDock = new VisuDock("view properties (debug) ", _viewer, this);
    _visuDock->setVisible(false);
    _visuDock->setFloating(true);
    this->addDockWidget(Qt::RightDockWidgetArea, _visuDock);

    setCentralWidget (_viewer);

    QAction * actionLoadMesh = new QAction("Load Mesh", this);
    QMenu * menuFile = new QMenu("File", this);
    menuFile->addAction(actionLoadMesh);
    connect(actionLoadMesh, SIGNAL(triggered()), this, SLOT(openMesh()));

    QAction * actionSaveMesh = new QAction("Save Mesh", this);
    menuFile->addAction(actionSaveMesh);
    connect(actionSaveMesh, SIGNAL(triggered()), this, SLOT(saveMesh()));

    QAction * actionLoadCamera = new QAction("Load Camera", this);
    menuFile->addAction(actionLoadCamera);
    connect(actionLoadCamera, SIGNAL(triggered()), this, SLOT(openCamera()));

    QAction * actionSaveCamera = new QAction("Save Camera", this);
    menuFile->addAction(actionSaveCamera);
    connect(actionSaveCamera, SIGNAL(triggered()), this, SLOT(saveCamera()));

    QAction * actionLoadSelections = new QAction("Load Selections", this);
    menuFile->addAction(actionLoadSelections);
    connect(actionLoadSelections, SIGNAL(triggered()), this, SLOT(loadSelections()));

    QAction * actionSaveSelections = new QAction("Save Selections", this);
    menuFile->addAction(actionSaveSelections);
    connect(actionSaveSelections, SIGNAL(triggered()), this, SLOT(saveSelections()));

    QAction * actionRecordVideo = new QAction("Record Video", this);
    menuFile->addAction(actionRecordVideo);
    connect(actionRecordVideo, SIGNAL(triggered()), this, SLOT(chooseVideoName()));

    QAction * actionLoadSVMselection = new QAction("Load SVM selection", this);
    menuFile->addAction(actionLoadSVMselection);
    connect(actionLoadSVMselection, SIGNAL(triggered()), this, SLOT(openSVMselection()));

    QAction * saveSelectionHist = new QAction("Save selection hist in image", this);
    menuFile->addAction(saveSelectionHist);
    connect(saveSelectionHist, SIGNAL(triggered()), _viewer, SLOT(saveSelHist()));

    QAction * exportTiming = new QAction("Export Timing", this);
    menuFile->addAction(exportTiming);
    connect(exportTiming, SIGNAL(triggered()), this, SLOT(saveTimings()));

    QAction * quitSimSel = new QAction("Quit ", this);
    menuFile->addAction(quitSimSel);
    quitSimSel->setShortcut(QKeySequence("Ctrl+Q"));
    connect(quitSimSel, SIGNAL(triggered()), this, SLOT(close()));

    QAction * actionHelp = new QAction("Help", this);
    actionHelp->setShortcut(QKeySequence("h"));
    QMenu * menuHelp = new QMenu("Help", this);
    menuHelp->addAction(actionHelp);
    connect(actionHelp, SIGNAL(triggered()), _viewer, SLOT(help()));


    QMenuBar * menubar = new QMenuBar(this);
    menubar->addAction(menuFile->menuAction());
    menubar->addAction(menuHelp->menuAction());
    this->setMenuBar(menubar);

    connect(_viewer, SIGNAL(signalTextToStatusBar(QString)), this, SLOT(setTextToStatusBar(QString)));
}

void MainWindow::saveMesh()
{
    QString selectedFilter, openFileNameLabel;
    QString fileFilter = "Known Filetypes (*.off);;OFF (*.off)";

    QString fileName = QFileDialog::getSaveFileName(this,
                                                    tr("Save the mesh"),
                                                    openFileNameLabel,
                                                    fileFilter,
                                                    &selectedFilter);


    statusBar()->showMessage("Saving mesh...");
    if(fileName.endsWith(".off") ){
        _viewer->saveMesh(fileName);

        statusBar()->showMessage("Mesh saved");
    }
}

void MainWindow::openMesh(){

    QString selectedFilter, openFileNameLabel;
    openFileNameLabel = "./models"; //"../../models"; //
    QString fileFilter = "Known Filetypes (*.off);;OFF (*.off);;OBJ (*.obj)";

    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Select an input mesh"),
                                                    openFileNameLabel,
                                                    fileFilter,
                                                    &selectedFilter);

    // In case of Cancel
    if ( fileName.isEmpty() ) {
        return;
    }

    statusBar()->showMessage("Opening mesh...");
    _viewer->open(fileName);

    statusBar()->showMessage("Mesh opened");

}

void MainWindow::openSVMselection(){

    QString selectedFilter, openFileNameLabel;
    openFileNameLabel = "../../../extern/libsvm-3.17/tools";
    QString fileFilter = "Known Filetypes (*.predict)";

    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Select an input SVM selection"),
                                                    openFileNameLabel,
                                                    fileFilter,
                                                    &selectedFilter);

    // In case of Cancel
    if ( fileName.isEmpty() ) {
        return;
    }

    statusBar()->showMessage("Opening SVM selection...");
    if(fileName.endsWith(".predict") ){
        _viewer->openSVMselection(fileName);

        statusBar()->showMessage("SVM selection opened");
    }
}


void MainWindow::setDebug(bool b){
    Trace(4, " setDebug SLOT ");
    _debugMode = b;
    _visuDock->setVisible(b);
}

void MainWindow::saveCamera(){
    QString fileName = QFileDialog::getSaveFileName(this, "Save camera file as ", "./data/", "Data (*.cam)");

    // In case of Cancel
    if ( fileName.isEmpty() )
        return;

    if(!fileName.endsWith(".cam"))
        fileName.append(".cam");

    _viewer->saveCamera(fileName);
    statusBar()->showMessage("Camera saved");
}

void MainWindow::saveTimings(){
    QString fileName = QFileDialog::getSaveFileName(this, "Save timing file as ", "./data/", "Data (*.txt)");

    // In case of Cancel
    if ( fileName.isEmpty() )
        return;

    if(!fileName.endsWith(".txt"))
        fileName.append(".txt");

    _viewer->saveTimings(fileName);
    statusBar()->showMessage("Timings saved");
}

void MainWindow::openCamera(){

    QString selectedFilter, openFileNameLabel;
    QString fileFilter = "Known Filetypes (*.cam);; Data (*.cam)";
    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Select an input camera"),
                                                    openFileNameLabel,
                                                    fileFilter,
                                                    &selectedFilter);
    // In case of Cancel
    if ( fileName.isEmpty() )
        return;

    _viewer->openCamera(fileName);
    statusBar()->showMessage("Camera loaded");
}

void MainWindow::saveSelections(){
    QString fileName = QFileDialog::getSaveFileName(this, "Save selections file as ", "./data/", "Data (*.sel)");

    // In case of Cancel
    if ( fileName.isEmpty() )
        return;

    if(!fileName.endsWith(".seg"))
        fileName.append(".seg");

    _viewer->saveSelections(fileName);
    statusBar()->showMessage("Selections saved");
}

void MainWindow::loadSelections(){

    QString selectedFilter, openFileNameLabel;
    QString fileFilter = "Known Filetypes (*.seg);; Data (*.seg)";
    QString fileName = QFileDialog::getOpenFileName(this,
                                                    tr("Select an input Selections file"),
                                                    openFileNameLabel,
                                                    fileFilter,
                                                    &selectedFilter);
    // In case of Cancel
    if ( fileName.isEmpty() )
        return;

    _viewer->loadSelections(fileName);
    statusBar()->showMessage("Selections loaded");
}

void MainWindow::chooseVideoName()
{
    QString selectedFilter, openFileNameLabel;
    QString fileFilter = "Known Filetypes (*.txt);; Data (*.txt)";
    QString fileName = QFileDialog::getSaveFileName(this ,
                                                    tr("Folder for the animation"),
                                                    openFileNameLabel,
                                                    fileFilter,
                                                    &selectedFilter);
    if ( fileName.isEmpty() ) {
        return;
    } else {
        _viewer->setVideoName(fileName);
        statusBar()->showMessage("Press z to start/end the video");
    }
}
void MainWindow::setTextToStatusBar(QString t)
{
    statusBar()->showMessage(t);
}
void MainWindow::closeEvent(QCloseEvent *event){
    Trace(1," close Event ");
    QMainWindow::closeEvent(event);
}
