/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "CotanLaplacian.h"

void CotanLaplacian::computeWeightAndAreaMaps(){

    Trace(2, "CotanLaplacian::computeWeightAndAreaMaps ");
    edgesWeightMap.clear();
    areaMap.clear();

    // compute wij = (cot(alpha(i,j))+cot(beta(i,j))/2
    for( unsigned int i = 0 ; i < triangles.size() ; i ++ )
    {
        const Triangle & triangle = triangles[i];

        Vec3Df p0 = (vertices[triangle.getVertex(0)]).getPos();
        Vec3Df p1 = (vertices[triangle.getVertex(1)]).getPos();
        Vec3Df p2 = (vertices[triangle.getVertex(2)]).getPos();

        float p0p1_slength = (p1-p0).getSquaredLength();
        float p1p2_slength = (p2-p1).getSquaredLength();
        float p2p0_slength = (p0-p2).getSquaredLength();

        float dot0 = Vec3Df::dotProduct(p1-p0,p2-p0);
        float dot1 = Vec3Df::dotProduct(p0-p1,p2-p1);
        float dot2 = Vec3Df::dotProduct(p0-p2,p1-p2);

        //cotan(p1pOp2)
        // =cos()/sin()
        // =cos()/sqrt(1-cos)
        // =dot0/sqrt(norm(p0p1)norm(p0p2)-dot0)
        float cot0 = dot0/sqrt(p0p1_slength*p2p0_slength-dot0*dot0);
        float cot1 = dot1/sqrt(p1p2_slength*p0p1_slength-dot1*dot1);
        float cot2 = dot2/sqrt(p1p2_slength*p2p0_slength-dot2*dot2);

        Edge edge0( triangle.getVertex(1) , triangle.getVertex(2) );
        Edge edge1( triangle.getVertex(0) , triangle.getVertex(2) );
        Edge edge2( triangle.getVertex(0) , triangle.getVertex(1) );

        EdgeWeights::iterator it = edgesWeightMap.find(edge0);
        if( it != edgesWeightMap.end() )
        {
            it->second = (it->second + cot0)/2.;
            areaMap[it->first] *= triangleArea(i);
        } else {
            edgesWeightMap[edge0] = cot0;
            areaMap[edge0] = triangleArea(i);
        }

        it = edgesWeightMap.find(edge1);
        if( it != edgesWeightMap.end() ){
            it->second = (it->second + cot1)/2.;
            areaMap[it->first] *= triangleArea(i);
        } else {
            edgesWeightMap[edge1] = cot1;
            areaMap[edge1] = triangleArea(i);
        }

        it = edgesWeightMap.find(edge2);
        if( it != edgesWeightMap.end() ){
            it->second = (it->second + cot2)/2.;
            areaMap[it->first] *= triangleArea(i);
        } else {
            edgesWeightMap[edge2] = cot2;
            areaMap[edge2] = triangleArea(i);
        }
    }
}

void CotanLaplacian::computeLaplacian()
{
    computeWeightAndAreaMaps();
    Trace(2,"CotanLaplacian::computeLaplacian");
    std::vector<std::pair<unsigned int,float> > laplacian_i;
    float sum;
    for( unsigned int i = 0; i < vertices.size() ; i++ )
    {
        laplacian_i.clear();
        sum = 0.f;
        for( unsigned int j = 0; j < oneRing[i].size() ; j++ )
        {
            Edge e( i, oneRing[i][j] );
            float wij = edgesWeightMap[e]/sqrt(areaMap[e]);
            laplacian_i.push_back(std::make_pair(oneRing[i][j],-wij) );
            sum += wij;
        }
        laplacian_i.push_back(std::make_pair(i,sum));
        laplacian.push_back(laplacian_i);
    }
}
