/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "GLUtilityMethods.h"
#include <cassert>

namespace GLCheck{
    int checkErrors( std::string const & szFile , int iLine )
    {
        int iRetCode;
        GLenum glErr = glGetError();
        while ( glErr != GL_NO_ERROR ) {
            std::cout << "GLError in file << " << szFile << " @ line " << iLine << " : " << gluErrorString( glErr ) << std::endl;
            iRetCode = 1;
            glErr = glGetError();
        }
        return iRetCode;
    }
}
namespace OpenGL{
    bool CheckExtension  ( std::string const & extName ) //( char * extName )
    {
        /*
         ** Search for extName in the extensions string.  Use of strstr()
         ** is not sufficient because extension names can be prefixes of
         ** other extension names.  Could use strtok() but the constant
         ** string returned by glGetString can be in read-only memory.
         */
        char *p = (char *) glGetString(GL_EXTENSIONS);
        char *end;
        int extNameLen;

        extNameLen =  extName.size();//strlen(extName);
        end = p + strlen(p);
    
        while (p < end) {
            int n = strcspn(p, " ");
            if ((extNameLen == n) && (extName.compare( 0, n, p) == 0)) {//(strncmp(extName, p, n) == 0)) {
                return GL_TRUE;
            }
            p += (n + 1);
        }
        return GL_FALSE;
    }


    bool CheckForErrors()
    {
        static bool checkForErrors = true;

        if( !checkForErrors )
        {
            return false;
        }

        const char * errorString = NULL;
        bool retVal = false;

        switch( glGetError() )
        {
        case GL_NO_ERROR:
            retVal = true;
            break;

        case GL_INVALID_ENUM:
            errorString = "GL_INVALID_ENUM";
            break;

        case GL_INVALID_VALUE:
            errorString = "GL_INVALID_VALUE";
            break;

        case GL_INVALID_OPERATION:
            errorString = "GL_INVALID_OPERATION";
            break;

        case GL_INVALID_FRAMEBUFFER_OPERATION:
            errorString = "GL_INVALID_FRAMEBUFFER_OPERATION";
            break;

        case GL_OUT_OF_MEMORY:
            errorString = "GL_OUT_OF_MEMORY";
            break;

        default:
            errorString = "UNKNOWN";
            break;
        }

        //
        if( !retVal )
        {
            printf( "%s\n", errorString );
            assert( retVal );
        }

        //
        return retVal;
    }

    //
    //
    void ClearErrors()
    {
        while( glGetError() != GL_NO_ERROR );
    }
}

namespace RGB
{
    int nColor = 30;

    float color[ 30 ][3] = {
        { 0.52459 , 0.370169 , 0.818006 } ,
        { 0.22797 , 0.736751 , 0.578866 } ,
        { 0.615946 , 0.840482 , 0.941527 } ,
        { 0.820386 , 0.998748 , 0.611765 } ,
        { 0.523095 , 0.277974 , 0.277974 } ,
        { 0.563622 , 0.205234 , 0.501854 } ,
        { 0.377066 , 0.299641 , 0.860731 } ,
        { 0.29955 , 0.770504 , 0.824872 } ,
        { 0.559197 , 0.574395 , 0.133989 } ,
        { 0.513832 , 0.308644 , 0.163775 } ,
        { 0.825589 , 0.496513 , 0.564584 } ,
        { 0.800473 , 0.423407 , 0.657435 } ,
        { 0.347524 , 0.813748 , 0.765499 } ,
        { 0.501244 , 0.87425 , 0.44152 } ,
        { 0.493843 , 0.905364 , 0.692546 } ,
        { 0.304311 , 0.506661 , 0.197818 } ,
        { 0.520516 , 0.124697 , 0.534661 } ,
        { 0.551827 , 0.655329 , 0.927062 } ,
        { 0.64477 , 0.907103 , 0.321981 } ,
        { 0.924224 , 0.835676 , 0.410391 } ,
        { 0.271122 , 0.30605 , 0.777478 } ,
        { 0.720943 , 0.366674 , 0.274281 } ,
        { 0.596414 , 0.334386 , 0.334386 } ,
        { 0.293492 , 0.801221 , 0.328527 } ,
        { 0.409415 , 0.607416 , 0.464027 } ,
        { 0.861845 , 0.425895 , 0.606302 } ,
        { 0.659052 , 0.409354 , 0.861982 } ,
        { 0.948363 , 0.724559 , 0.358358 } ,
        { 0.460586 , 0.399985 , 0.996397 } ,
        { 0.49691 , 0.196338 , 0.592523 }
    };

//    float color_emilie[ 20 ][3] = {
//        { 0.152941176 ,0.521568627 , 1.0 } ,
//        { 0.14117647058 , 0.90980392156 , 0.43137254902 } ,
//        { 1.000 , 0.96 , 0.16470588235 } ,
//        { 0.90980392156 , 0.54509803921 , 0.14117647058 } ,
//        { 1.000 , 0.15294117647 , 0.46274509803 } ,

//        { 0.563622 , 0.205234 , 0.501854 } ,
//        { 0.820386 , 0.886748 , 0.611765 } ,
//        { 0.523095 , 0.277974 , 0.277974 } ,
//        { 0.948363 , 0.724559 , 0.358358 } ,
//        { 0.513832 , 0.308644 , 0.163775 } ,
//        { 0.825589 , 0.496513 , 0.564584 } ,
//        { 0.863622 , 0.205234 , 0.501854 } ,
//        { 0.347524 , 0.813748 , 0.765499 } ,
//        { 0.501244 , 0.87425  , 0.44152 } ,
//        { 0.304311 , 0.506661 , 0.197818 } ,
//        { 0.520516 , 0.124697 , 0.534661 } ,
//        { 0.551827 , 0.655329 , 0.927062 } ,
//        { 0.64477 , 0.907103  , 0.321981 } ,
//        { 0.924224 , 0.835676 , 0.410391 } ,
//        { 0.271122 , 0.30605 , 0.777478 } ,
//        { 0.720943 , 0.366674 , 0.274281 } ,
//        { 0.596414 , 0.334386 , 0.334386 } ,
//        { 0.293492 , 0.801221 , 0.328527 } ,
//        { 0.409415 , 0.607416 , 0.464027 } ,
//        { 0.861845 , 0.425895 , 0.606302 } ,
//        { 0.49691 , 0.196338 , 0.592523 }
//    };

    int colorSelection[16][3] = {
        //
       {205, 1, 1},
        { 241 , 184 ,  91} ,

        { 88 , 2 , 140} ,
        { 255 , 104 , 34} ,
        { 182 , 247 , 21} ,
        { 137 , 11 , 194},
        { 200 , 2 , 44} ,
        { 96 , 76 , 219 } ,
        { 168 , 104 ,219} ,
        { 142 , 144 , 34 } ,

        {131,  78, 41},
        { 195 , 100 , 19} ,
        { 2 , 172 , 200},

        { 39 , 133 , 255} ,
        { 20 , 250 , 45} ,
        { 76 ,196  , 210}
    };

    float color4[ 30 ][4] = {
        { 0.52459 , 0.370169 , 0.818006 , 1.f } ,
        { 0.22797 , 0.736751 , 0.578866 , 1.f } ,
        { 0.615946 , 0.840482 , 0.941527 , 1.f } ,
        { 0.820386 , 0.886748 , 0.611765 , 1.f } ,
        { 0.523095 , 0.277974 , 0.277974 , 1.f } ,
        { 0.160586 , 0.399985 , 0.656397 , 1.f } ,
        { 0.563622 , 0.205234 , 0.501854 , 1.f } ,
        { 0.948363 , 0.724559 , 0.358358 , 1.f } ,
        { 0.377066 , 0.299641 , 0.860731 , 1.f } ,
        { 0.29955 , 0.770504 , 0.824872 , 1.f } ,
        { 0.659052 , 0.409354 , 0.861982 , 1.f } ,
        { 0.559197 , 0.574395 , 0.133989 , 1.f } ,
        { 0.513832 , 0.308644 , 0.163775 , 1.f } ,
        { 0.825589 , 0.496513 , 0.564584 , 1.f } ,
        { 0.800473 , 0.423407 , 0.657435 , 1.f } ,
        { 0.347524 , 0.813748 , 0.765499 , 1.f } ,
        { 0.501244 , 0.87425 , 0.44152 , 1.f } ,
        { 0.493843 , 0.905364 , 0.692546 , 1.f } ,
        { 0.304311 , 0.506661 , 0.197818 , 1.f } ,
        { 0.520516 , 0.124697 , 0.534661 , 1.f } ,
        { 0.551827 , 0.655329 , 0.927062 , 1.f } ,
        { 0.64477 , 0.907103 , 0.321981 , 1.f } ,
        { 0.924224 , 0.835676 , 0.410391 , 1.f } ,
        { 0.271122 , 0.30605 , 0.777478 , 1.f } ,
        { 0.720943 , 0.366674 , 0.274281 , 1.f } ,
        { 0.596414 , 0.334386 , 0.334386 , 1.f } ,
        { 0.293492 , 0.801221 , 0.328527 , 1.f } ,
        { 0.409415 , 0.607416 , 0.464027 , 1.f } ,
        { 0.861845 , 0.425895 , 0.606302 , 1.f } ,
        { 0.49691 , 0.196338 , 0.592523 , 1.f }
    };


    void get_random_RGB( float & r , float & g , float & b )
    {
        r = (float)(rand()) / (float)(RAND_MAX);
        g = (float)(rand()) / (float)(RAND_MAX);
        b = (float)(rand()) / (float)(RAND_MAX);
    }

    void get_random_RGB_from_HSV( float & r , float & g , float & b )
    {
        QColor c = QColor::fromHsvF( (float)(rand()) / (float)(RAND_MAX) , 0.3f + 0.5f * (float)(rand()) / (float)(RAND_MAX) , 0.5f + 0.5f * (float)(rand()) / (float)(RAND_MAX) );

        r = c.redF();
        g = c.greenF();
        b = c.blueF();
    }

    void get_random_RGB_from_HSV( float & r , float & g , float & b , float H )
    {
        QColor c = QColor::fromHsvF( H , 0.3f + 0.5f * (float)(rand()) / (float)(RAND_MAX) , 0.5f + 0.5f * (float)(rand()) / (float)(RAND_MAX) );

        r = c.redF();
        g = c.greenF();
        b = c.blueF();
    }

    void calc_RGB( float val , float val_min , float val_max , float & r , float & g , float & b )
    {
        // define uniform color intervalls [v0,v1,v2,v3,v4]
        float v0, v1, v2, v3, v4 ;

        v0 = val_min ;
        v1 = val_min + 1.0/4.0 * (val_max - val_min);
        v2 = val_min + 2.0/4.0 * (val_max - val_min);
        v3 = val_min + 3.0/4.0 * (val_max - val_min);
        v4 = val_max ;



        if (val < v0)
        {
            r = 0.f;
            g = 0.f;
            b = 1.f;
            return;
        }
        else if (val > v4)
        {
            r = 1.f;
            g = 0.f;
            b = 0.f;
            return;
        }
        else if (val <= v2)
        {
            if (val <= v1) // [v0, v1]
            {
                r = 0.f;
                g = (val - v0) / (v1 - v0);
                b = 1.f;
                return;
            }
            else // ]v1, v2]
            {
                r = 0.f;
                g = 1.f;
                b = 1.f - (val - v1) / (v2 - v1);
                return;
            }
        }
        else
        {
            if (val <= v3) // ]v2, v3]
            {
                r = (val - v2) / (v3 - v2);
                g = 1.f;
                b = 0.f;
                return;
            }
            else // ]v3, v4]
            {
                r = 1.f;
                g = 1.f - (val - v3) / (v4 - v3);
                b = 0.f;
                return;
            }
        }
    }
}










namespace BasicGL
{
    void drawSphere(float x,float y,float z,float radius,int slices,int stacks)
    {
        if(stacks < 2){stacks = 2;}
        if(stacks > 20){stacks = 20;}
        if(slices < 3){slices = 3;}
        if(slices > 30){slices = 30;}
        //Pas essentiel ...

        int Nb = slices*stacks +2;
        std::vector< qglviewer::Vec > points(Nb);

        qglviewer::Vec centre(x,y,z);

        float sinP , cosP , sinT , cosT , Phi , Theta;
        points[0] = qglviewer::Vec(0,0,1);
        points[Nb-1] = qglviewer::Vec(0,0,-1);

        for(int i=1; i<=stacks; i++)
        {
            Phi = 90 - (float)(i*180)/(float)(stacks+1);
            sinP = sinf(Phi*3.14159265/180);
            cosP = cosf(Phi*3.14159265/180);

            for(int j=1; j<=slices; j++)
            {
                Theta = (float)(j*360)/(float)(slices);
                sinT = sinf(Theta*3.14159265/180);
                cosT = cosf(Theta*3.14159265/180);

                points[ j + (i-1)*slices ] = qglviewer::Vec(cosT*cosP,sinT*cosP,sinP);
            }
        }

        int k1,k2;
        glBegin(GL_TRIANGLES);
        for(int i=1; i<=slices; i++)
        {
            k1 = i;
            k2 = (i%slices+1);
            glNormal3fv(points[0]);
            glVertex3fv((centre + radius*points[0]));
            glNormal3fv(points[k1]);
            glVertex3fv((centre + radius*points[k1]));
            glNormal3fv(points[k2]);
            glVertex3fv((centre + radius*points[k2]));

            k1 = (stacks-1)*slices+i;
            k2 = (stacks-1)*slices+(i%slices+1);
            glNormal3fv(points[k1]);
            glVertex3fv((centre + radius*points[k1]));
            glNormal3fv(points[Nb-1]);
            glVertex3fv((centre + radius*points[Nb-1]));
            glNormal3fv(points[k2]);
            glVertex3fv((centre + radius*points[k2]));
        }
        glEnd();

        glBegin(GL_QUADS);
        for(int j=1; j<stacks; j++)
        {
            for(int i=1; i<=slices; i++)
            {
                k1 = i + (j-1)*slices;
                k2 = (i%slices+1) + (j-1)*slices;
                glNormal3fv(points[k2]);
                glVertex3fv((centre + radius*points[k2]));
                glNormal3fv(points[k1]);
                glVertex3fv((centre + radius*points[k1]));

                k1 = i + (j)*slices;
                k2 = (i%slices+1) + (j)*slices;
                glNormal3fv(points[k1]);
                glVertex3fv((centre + radius*points[k1]));
                glNormal3fv(points[k2]);
                glVertex3fv((centre + radius*points[k2]));
            }
        }
        glEnd();
    }
}

namespace Math{
    // ---------------------------------------------------------------------------------
    // get a random number (int)
    // ---------------------------------------------------------------------------------
    int	getRandomIntBetween(int iMin, int iMax)
    {
            const float fDiv = 1.0f / (RAND_MAX + 1.0f);
            int iRand = rand();
            int iRange = iMax - iMin + 1;
            return iMin + (int) ( iRange * iRand * fDiv );
    }

    // ---------------------------------------------------------------------------------
    // get a random number (float)
    // ---------------------------------------------------------------------------------
    float getRandomFloatBetween(float fMin, float fMax)
    {
            const float fDiv = 1.0f / (RAND_MAX + 1.0f);
            int iRand = rand();
            float fRange = fMax - fMin + 1.0f;
            return fMin + ( fRange * iRand * fDiv );
    }
}
