/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "SVMclassification.h"
#define Malloc(type,n) (type *)malloc((n)*sizeof(type))

SVMClassification::SVMClassification(){
    // default values
    param.svm_type = C_SVC;
    param.kernel_type = RBF;
    param.degree = 3;
    param.gamma = 0;	// 1/num_features
    param.coef0 = 0;
    param.nu = 0.5;
    param.cache_size = 100;
    param.C = 1;
    param.eps = 1e-3;
    param.p = 0.1;
    param.shrinking = 1;
    param.probability = 0;
    param.nr_weight = 0;
    param.weight_label = NULL;
    param.weight = NULL;

    model_allocated = false;
    param_allocated = false;
    x_space_allocated = false;
    prob_allocated = false;
}

// always set parameters before problem.
void SVMClassification::setSVMproblem(std::vector<int> & _labels,
                                      std::vector<std::vector<double> > & _vectors)
{
    Trace(1, "setSVMproblems");
    assert(!_vectors.empty());

    prob.l = _labels.size();

    prob.y = Malloc(double,prob.l);
    prob.x = Malloc(struct svm_node *,prob.l);
    prob_allocated = true;

    x_space = Malloc(struct svm_node, _vectors.size()*(_vectors[0].size()+1));
    x_space_allocated = true;

    int max_index = 0;
    unsigned int j = 0;
    for(unsigned int i=0; i < (unsigned int) prob.l; i++)
    {
        prob.y[i] = _labels[i];

        prob.x[i] = &x_space[j];
        for (unsigned int k = 0; k < _vectors[i].size(); k++)
        {
            if(param.kernel_type != PRECOMPUTED)
                x_space[j].index = (int) k+1;
            else
                x_space[j].index = (int) k;

            max_index = std::max(max_index, x_space[j].index);
            x_space[j].value = _vectors[i][k];
            j++;
        }
        x_space[j++].index = -1;
    }

    param_allocated = true;
    if(param.gamma == 0 && max_index > 0)
        param.gamma = 1.0/ max_index;


    if(param.kernel_type == PRECOMPUTED)
    {
        for (int i=0; i < prob.l; i++)
        {
            if (prob.x[i][0].index != 0)
            {
                fprintf(stderr,"Wrong input format: first column must be 0:sample_serial_number\n");
                exit(1);
            }
            if ((int)prob.x[i][0].value <= 0 || (int)prob.x[i][0].value > max_index)
            {
                fprintf(stderr,"Wrong input format: sample_serial_number out of range\n");
                exit(1);
            }
        }
    }
}

void SVMClassification::setSVMparameters(int _svm_type,
                                         int _kernel_type,
                                         int _degree,
                                         double _gamma, double _coef0 ){
    Trace(1, "setSVMparameters");
    param.svm_type = _svm_type;
    param.kernel_type = _kernel_type;
    param.degree = _degree;
    param.gamma = _gamma;
    param.coef0 = _coef0;
    param_allocated = true;
}

    //todo gerer predict_probability
void SVMClassification::predict(std::vector<int> & _labels,
                                std::vector<std::vector<double> > & _vectors){
//    int correct = 0;
//    int total = 0;
//    double error = 0;
//    double sump = 0, sumt = 0, sumpp = 0, sumtt = 0, sumpt = 0;

    struct svm_node* datas;
    datas = NULL;
    for (unsigned int i = 0; i < _vectors.size(); i++)
    {
        datas =  (struct svm_node *) realloc(datas, (_vectors[i].size()+1)*sizeof(struct svm_node));

        for (unsigned int j = 0; j < _vectors[i].size(); j++)
        {
        if (param.kernel_type != PRECOMPUTED)
            datas[j].index = (int) j+1;
        else
            datas[j].index = (int) j;
        datas[j].value = _vectors[i][j];
        }
        datas[_vectors[i].size()].index = -1;
        _labels[i] = svm_predict( model, datas);
        Trace(4, "label "<<i<<":" << _labels[i]);

//        if(predict_label == target_label)
//            ++correct;
//        error += (predict_label-target_label)*(predict_label-target_label);
//        sump += predict_label;
//        sumt += target_label;
//        sumpp += predict_label*predict_label;
//        sumtt += target_label*target_label;
//        sumpt += predict_label*target_label;
//        ++total;
    }
}

void SVMClassification::predictWithProb (std::vector<int> & _labels,
                                         std::vector<std::vector<double> > & _vectors,
                                         std::vector< std::vector<double> > & probabilities){
//    int correct = 0;
//    int total = 0;
//    double error = 0;
//    double sump = 0, sumt = 0, sumpp = 0, sumtt = 0, sumpt = 0;
    if(svm_check_probability_model(model)==0)
    {
        fprintf(stderr,"Model does not support probabiliy estimates\n");
        return;
    }
    struct svm_node* datas;
    datas = NULL;
    probabilities.clear();
    std::vector<double>  probabilities_i;

    int nr_class=svm_get_nr_class(model);

    for (unsigned int i = 0; i < _vectors.size(); i++)
    {
        datas =  (struct svm_node *) realloc(datas, (_vectors[i].size()+1)*sizeof(struct svm_node));

        double *prob_estimates=NULL;
        prob_estimates = (double *) malloc(nr_class*sizeof(double));

        for (unsigned int j = 0; j < _vectors[i].size(); j++)
        {
        if (param.kernel_type != PRECOMPUTED)
            datas[j].index = (int) j+1;
        else
            datas[j].index = (int) j;
        datas[j].value = _vectors[i][j];
        }
        datas[_vectors[i].size()].index = -1;
        _labels[i] = svm_predict_probability( model, datas, prob_estimates);
        probabilities_i.clear();
        for (int j = 0; j < nr_class; j++)
        {
            probabilities_i.push_back(prob_estimates[j]);
        }
        probabilities.push_back(probabilities_i);
       Trace(4, "label "<<i<<":" << _labels[i]);
        free(prob_estimates);

//        if(predict_label == target_label)
//            ++correct;
//        error += (predict_label-target_label)*(predict_label-target_label);
//        sump += predict_label;
//        sumt += target_label;
//        sumpp += predict_label*predict_label;
//        sumtt += target_label*target_label;
//        sumpt += predict_label*target_label;
//        ++total;
    }
    free(datas);
    Trace(1, "probalities_size " << probabilities.size() << "x"<<probabilities[0].size() );

}

void SVMClassification::do_cross_validation()
{
    Trace(1, "SVMClassification::do_cross_validation");
    int i;
    int total_correct = 0;
    double total_error = 0;
    double sumv = 0, sumy = 0, sumvv = 0, sumyy = 0, sumvy = 0;
    double *target = Malloc(double,prob.l);

    svm_cross_validation(&prob,&param,nr_fold,target);
    if(param.svm_type == EPSILON_SVR ||
       param.svm_type == NU_SVR)
    {
        for(i=0;i<prob.l;i++)
        {
            double y = prob.y[i];
            double v = target[i];
            total_error += (v-y)*(v-y);
            sumv += v;
            sumy += y;
            sumvv += v*v;
            sumyy += y*y;
            sumvy += v*y;
        }
        printf("Cross Validation Mean squared error = %g\n",total_error/prob.l);
        printf("Cross Validation Squared correlation coefficient = %g\n",
            ((prob.l*sumvy-sumv*sumy)*(prob.l*sumvy-sumv*sumy))/
            ((prob.l*sumvv-sumv*sumv)*(prob.l*sumyy-sumy*sumy))
            );
    }
    else
    {
        for(i=0;i<prob.l;i++)
            if(target[i] == prob.y[i])
                ++total_correct;
        printf("Cross Validation Accuracy = %g%%\n",100.0*total_correct/prob.l);
    }
    free(target);
}

void SVMClassification::train(){
    Trace(1, "SVMClassification::train");
    model = svm_train(&prob,&param);
    model_allocated = true;
}

void SVMClassification::saveModelInFile(std::string model_file_name)
{
    Trace(1, "saveModelInFile");
    const char* file_name = model_file_name.c_str();
    if(svm_save_model(file_name,model))
    {
        fprintf(stderr, "can't save model to file %s\n", file_name);
        exit(1);
    }
}

void SVMClassification::clear(){
    svm_destroy_param(&param);
    if (model_allocated)
        svm_free_and_destroy_model(&model);
    if (prob_allocated)
    {
        free(prob.y);
        free(prob.x);
    }
    if (x_space_allocated)
        free(x_space);
    model_allocated = false;
    param_allocated = false;
    x_space_allocated = false;
    prob_allocated = false;
}
